@@ -10,20 +10,20 @@ use B::Hooks::OP::PPAddr;
 use Devel::Declare::Context::Simple;
 use Parse::Method::Signatures;
 use Moose::Util::TypeConstraints;
-use Scope::Upper qw/unwind want_at :words/;
-use TryCatch::Exception;
+use Scope::Upper qw/localize unwind want_at :words/;
 use Carp qw/croak/;
+use XSLoader;
 
-use base 'DynaLoader';
+use base qw/Devel::Declare::Context::Simple/;
 
 
-our $VERSION = '1.000003';
-our $PARSE_CATCH_NEXT = 0;
+our $VERSION = '1.001000';
+
+# These are private state variables. Mess with them at your peril
 our ($CHECK_OP_HOOK, $CHECK_OP_DEPTH) = (undef, 0);
+our (@STATE);
 
-sub dl_load_flags { 0x01 }
-
-__PACKAGE__->bootstrap($VERSION);
+XSLoader::load(__PACKAGE__, $VERSION);
 
 use namespace::clean;
 
@@ -40,9 +40,6 @@ use Sub::Exporter -setup => {
           { $name => { const => sub { $parser->($pack, @_) } } },
         );
       }
-      if (my $code = __PACKAGE__->can("_extras_for_${name}")) {
-        $code->($pack);
-      }
     }
     Sub::Exporter::default_installer(@_);
 
@@ -51,34 +48,25 @@ use Sub::Exporter -setup => {
 
 
 # The actual try call itself. Nothing to do with parsing.
-sub try {
-  my ($sub) = @_;
-
-  return new TryCatch::Exception( try => $sub, ctx => SUB(CALLER(1)) );
-
+sub try () {
+  warn "non-shadowed - shouldn't happen";
+  return;
 }
 
 # Where we store all the TCs for catch blocks created at compile time
 # Not sure we really want to do this, but we will for now.
 our $TC_LIBRARY = {};
 
-sub get_tc {
+sub check_tc {
   my ($class, $tc) = @_;
 
-  $TC_LIBRARY->{$tc} or die "Unable to find parse TC for '$tc'";
+  my $type = $TC_LIBRARY->{$tc} or die "Unable to find parse TC for '$tc'";
+
+  return $type->check($TryCatch::Error);
 }
 
 # From here on out its parsing methods.
 
-sub _extras_for_try {
-  my ($pack) = @_;
-
-  Devel::Declare->setup_for(
-    $pack,
-    { catch => { const => sub { _parse_catch($pack, @_) } } }
-  );
-}
-
 # Replace 'try {' with an 'try (sub {'
 sub _parse_try {
   my $pack = shift;
@@ -86,43 +74,66 @@ sub _parse_try {
   # Hide Devel::Declare from carp;
   local $Carp::Internal{'Devel::Declare'} = 1;
 
-  my $ctx = Devel::Declare::Context::Simple->new->init(@_);
+  my $ctx = TryCatch->new->init(@_);
 
-  if (my $len = Devel::Declare::toke_scan_ident( $ctx->offset )) {
-    $ctx->inc_offset($len);
-    $ctx->skipspace;
+  $ctx->skip_declarator;
+  $ctx->skipspace;
 
-    my $linestr = $ctx->get_linestr;
-    croak "block required after try"
-      unless substr($linestr, $ctx->offset, 1) eq '{';
+  my $linestr = $ctx->get_linestr;
 
-    substr($linestr, $ctx->offset+1,0) = q# BEGIN { TryCatch::postlude() }#;
-    substr($linestr, $ctx->offset,0) = q#(sub #;
-    $ctx->set_linestr($linestr);
+  # Let "try =>" be valid.
+  return if substr($linestr, $ctx->offset, 2) eq '=>';
 
-    if (! $CHECK_OP_DEPTH++) {
-      $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
-    }
+  # Shadow try to be a constant no-op sub
+  $ctx->shadow(sub () { } );
 
+  $ctx->inject_if_block(
+    $ctx->injected_try_code . $ctx->scope_injector_call,
+    q#;#
+  ) or croak "block required after try";
+
+  #$ctx->debug_linestr("try");
+  if (! $CHECK_OP_DEPTH++) {
+    $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
   }
+
+  $ctx->debug_linestr('post try');
+
+  # Number of catch blocks found, we only care about 0,1 and +1 cases tho
+  push @STATE, 0;
   
 }
 
-sub prelude {
-  my ($ctx, $linestr) = @_;
+sub injected_try_code {
+  # try { ...
+  # ->
+  # try; { local $@; eval { ...
 
-  substr($linestr, $ctx->offset+1, 0, 
-         "BEGIN { TryCatch::postlude }; local *_{ARRAY}");
+  return @STATE > 1
+       ? 'local $TryCatch::CTX = Scope::Upper::HERE; eval {' # Nested case
+       : 'local $@; eval {'
 }
 
-sub postlude {
-  on_scope_end { block_postlude() }
+sub injected_after_try {
+  # This semicolon is for the end of the eval
+  return '; if ($@) { local $TryCatch::Error = $@;';
+}
+
+sub injected_no_catch_code {
+  return "};";
+}
+
+sub inject_scope {
+  on_scope_end { 
+    block_postlude() 
+  }
 }
 
 # Called after the block from try {} or catch {}
+# Look ahead and determine what action to take based on wether or note we
+# see a 'catch' token after the block
 sub block_postlude {
-
-  my $ctx = Devel::Declare::Context::Simple->new->init(
+  my $ctx = TryCatch->new->init(
     '', 
     Devel::Declare::get_linestr_offset()
   );
@@ -133,22 +144,36 @@ sub block_postlude {
   my $toke = '';
   my $len = 0;
 
+  # Since we're not being called from a normal D::D callback, we have to
+  # find this info manually.
   if ($len = Devel::Declare::toke_scan_word($offset, 1 )) {
     $toke = substr( $linestr, $offset, $len );
+    $ctx->{Declarator} = $toke;
   }
 
-  if (--$CHECK_OP_DEPTH == 0) {
+  if ($CHECK_OP_DEPTH && --$CHECK_OP_DEPTH == 0) {
     TryCatch::XS::uninstall_return_op_check($CHECK_OP_HOOK);
   }
-  if ($toke eq 'catch') {
 
-    $ctx->skipspace;
-    substr($linestr, $ctx->offset, 0) = ')->';
-    $ctx->set_linestr($linestr);
-    $TryCatch::PARSE_CATCH_NEXT = 1;
-  } else {
-    substr($linestr, $offset, 0) = ')->run(@_);';
+  if ($toke eq 'catch') {
+    # We don't want the 'catch' token in the output since it messes up the
+    # if/else we build up. So dont let control go back to perl just yet.
+
+    $ctx->_parse_catch;
+
+  } else  {
+    my $code = $STATE[-1] == 0
+             ? $ctx->injected_no_catch_code
+             : '}}';
+
+    substr($linestr, $offset, 0, $code);
+
+
     $ctx->set_linestr($linestr);
+    $ctx->debug_linestr("finalizer");
+
+    # This try/catch stmt is finished
+    pop @STATE;
   }
 }
 
@@ -156,36 +181,36 @@ sub block_postlude {
 # turn 'catch() {' into '->catch({ TC_check_code;'
 # the '->' is added by one of the postlude hooks
 sub _parse_catch {
-  my $pack = shift;
-  my $ctx = Devel::Declare::Context::Simple->new->init(@_);
+  my $ctx = shift;
+  my $pack = $ctx->get_curstash_name;
 
-  return unless $TryCatch::PARSE_CATCH_NEXT;
-  $TryCatch::PARSE_CATCH_NEXT = 0;
 
   # Hide Devel::Declare from carp;
   local $Carp::Internal{'Devel::Declare'} = 1;
   local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
   local $Carp::Internal{'TryCatch'} = 1;
 
+  # This isn't a normal DD-callback, so we can strip_name to get rid of try
+  my $offset = $ctx->offset;
+  $ctx->strip_name;
   $ctx->skipspace;
+  my $new_offset = $ctx->offset;
+ 
+  $ctx->debug_linestr('catch');
   my $linestr = $ctx->get_linestr;
 
-  my $len = length "->catch";
-  my $sub = substr($linestr, $ctx->offset, $len);
-  die "_parse_catch expects to find '->catch' in linestr, found: "  
-    . substr($linestr, $ctx->offset, $len)
-    unless $sub eq '->catch';
-
-  $ctx->inc_offset($len);
-  $ctx->skipspace;
-
+  my $code;
   my $var_code = "";
+
   my @conditions;
   # optional ()
   if (substr($linestr, $ctx->offset, 1) eq '(') {
-    my $substr = substr($linestr, $ctx->offset+1);
+    my $proto = $ctx->strip_proto;
+    croak "Run-away catch signature"
+      unless (length $proto);
+    
     my $sig = Parse::Method::Signatures->new(
-      input => $substr,
+      input => $proto,
       from_namespace => $pack,
     );
     my $errctx = $sig->ppi;
@@ -196,49 +221,69 @@ sub _parse_catch {
 
     my $left = $sig->remaining_input;
 
-    die "can't handle un-named vars yet" unless $param->can('variable_name');
+    croak "TryCatch can't handle un-named vars in catch signature" 
+      unless $param->can('variable_name');
 
     my $name = $param->variable_name;
-    $var_code .= "my $name = \$@;";
+    $var_code = "my $name = \$TryCatch::Error;";
 
     # (TC $var)
     if ($param->has_type_constraints) {
       my $tc = $param->meta_type_constraint;
       $TC_LIBRARY->{"$tc"} = $tc;
-      push @conditions, "'$tc'";
+      push @conditions, "TryCatch->check_tc('$tc')";
     }
 
     # ($var where { $_ } )
     if ($param->has_constraints) {
       foreach my $con (@{$param->constraints}) {
-        push @conditions, "sub $con";
+        $con =~ s/^{|}$//g;
+        push @conditions, "do {local \$_ = \$TryCatch::Error; $con }";
       }
     }
 
-    substr($linestr, $ctx->offset, length($linestr) - $ctx->offset - length($left), '');
-    $ctx->set_linestr($linestr);
-    $ctx->skipspace;
-    if (substr($linestr, $ctx->offset, 1) ne ')') {
-      croak "')' expected after catch signature";
-    }
-
-    substr($linestr, $ctx->offset, 1, '');
-    $ctx->set_linestr($linestr);
-    $ctx->skipspace;
+    $linestr = $ctx->get_linestr;
   }
 
-  croak "block required after catch"
-    unless substr($linestr, $ctx->offset, 1) eq '{';
+  $code = $ctx->injected_after_try
+    if $STATE[-1] == 0;
 
-  substr($linestr, $ctx->offset+1,0) = 
-    q# BEGIN { TryCatch::postlude() }# . $var_code;
-  push @conditions, "sub ";
-  substr($linestr, $ctx->offset,0) = '(' . join(', ', @conditions);
-  $ctx->set_linestr($linestr);
+  @conditions = ('1')
+    unless @conditions;
 
-  if (! $CHECK_OP_DEPTH++) {
-    $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
-  }
+  $code .= $STATE[-1] < 1
+         ? "if ("
+         : "elsif (";
+
+  $ctx->inject_if_block(
+    $ctx->scope_injector_call . $var_code,
+    $code . join(' && ', @conditions) . ')'
+  ) or croak "block required after catch";
+
+  $ctx->debug_linestr('post catch');
+
+  #if (! $CHECK_OP_DEPTH++) {
+  #  $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
+  #}
+
+  $STATE[-1]++;
+}
+
+sub debug_linestr {
+  my ($ctx, $message) = @_;
+
+  return unless $ENV{TRYCATCH_DEBUG};
+ 
+  local $Carp::Internal{'TryCatch'} = 1;
+  local $Carp::Internal{'Devel::Declare'} = 1;
+  local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
+  local $Carp::Internal{'Devel::PartialDump'} = 1;
+  Carp::cluck($message) if $message;
+
+  require Devel::PartialDump;
+
+  warn   "  Substr: ", Devel::PartialDump::dump(substr($ctx->get_linestr, $ctx->offset)),
+       "\n  Whole:  ", Devel::PartialDump::dump($ctx->get_linestr), "\n\n";
 }
 
 
